import requests
import yaml
import hashlib
from deepmerge import always_merger

CFG_FILE = 'timona.yaml'


def load_config_file(c):
    _data = []

    def _load_config(c):
        if c.startswith('http://') or c.startswith('https://'):
            r = requests.get(c, timeout=(0.5, 5))
            r.raise_for_status()
            cfg = r.text
        else:
            with open(c) as f:
                cfg = f.read()
        for ln in cfg.split('\n'):
            ln = ln.rstrip()
            if ln.startswith('#+'):
                _load_config(ln[2:])
            elif ln:
                _data.append(ln)
    _load_config(c)
    _data = '\n'.join(_data)
    return _data


def load_config(c=None):
    if c is None:
        c = CFG_FILE
    _data = {}

    def _load_config(c, _data):
        f = load_config_file(c)
        f = yaml.safe_load(f)
        if 'include' in f:
            for g in f['include']:
                _load_config(g, _data)
            del(f['include'])
        _data = always_merger.merge(_data, f)  # _data.update(f)
    _load_config(c, _data)
    h = hashlib.sha256(
        yaml.safe_dump(_data).encode('utf-8')
    ).hexdigest()  # sha256(yaml(_data))
    _data['.hash'] = h
    return _data



#!/usr/bin/env python3

import yaml
import sys
import os

from .__version__ import __version__ as version
from .config import load_config
from .template import Template
from .helm import Helm
from .releases import get_releases


def usage():
    print()
    print("timona COMAND [RELEASE]")
    print()
    print("COMMANDS:")
    print("  config           - dump config")
    print("  releases         - list releases")
    print("  env RELEASE      - show environment")
    print("  template RELEASE - render template")
    print("  debug RELEASE    - dry-run")
    print("  diff RELEASE     - diff between running and config")
    print("  deploy RELEASE   - deploy")
    print("  status RELEASE   - status")
    print("  delete RELEASE   - delete")
    print("  version          - show version(s)")
    print()
    print("https://github.com/mihaiush/timona/wiki")
    print()
    sys.exit(2)


def main():
    if len(sys.argv) == 1:
        usage()
    CMD = sys.argv[1]
    valid_commands = [
        'config',
        'releases',
        'env',
        'template',
        'debug',
        'diff',
        'deploy',
        'status',
        'delete',
        'version'
    ]
    if CMD not in valid_commands:
        usage()
    R = None
    if CMD not in ['config', 'releases', 'version']:
        if len(sys.argv) == 2:
            usage()
        R = sys.argv[2]

    CONFIG = load_config()

    if CMD == 'config':
        print(yaml.safe_dump(CONFIG))
        sys.exit()

    TMP = '{}/.timona'.format(CONFIG['config']['tmp'])
    os.makedirs(TMP, exist_ok=True)
    T = Template(CONFIG['config']['template'], TMP)
    H = Helm(CONFIG['config']['helm'], T, TMP)

    if CMD == 'version':
        print('timona: {}'.format(version))
        print('template: {}'.format(T.version))
        print('helm: {}'.format(H.version))
        print('helm: diff: {}'.format(H.version_diff))
        sys.exit()

    RELEASES = get_releases(T, TMP, CONFIG)

    if R and R not in RELEASES:
        raise RuntimeError('Unknown release: {}'.format(R))

    def env():
        print()
        e = {R: {'values': VALUES, 'variables': RELEASES[R]}}
        print(yaml.safe_dump(e))

    if CMD == 'releases':
        print()
        for r in RELEASES.keys():
            print('  {}'.format(r))
        print()
        sys.exit()

    VALUES = T.render(CONFIG['values'], RELEASES[R]).strip()

    if CMD == 'env':
        env()
        sys.exit()

    if CMD == 'template':
        H.template(R, RELEASES[R], VALUES)
        sys.exit()

    if CMD == 'debug':
        env()
        H.debug(R, RELEASES[R], VALUES)
        sys.exit()

    if CMD == 'diff':
        H.diff(R, RELEASES[R], VALUES)
        sys.exit()

    if CMD == 'deploy':
        H.deploy(R, RELEASES[R], VALUES)
        sys.exit()

    if CMD == 'status':
        env()
        H.status(R, RELEASES[R])
        sys.exit()

    if CMD == 'delete':
        H.delete(R, RELEASES[R])
        sys.exit()



import subprocess
import shlex


class Helm():

    def __init__(self, config, tpl, tmp):
        self.cmd = '{} {}'.format(config['command'], config['flags'])
        if 'wait' in config and config['wait']:
            self.atomic = '--atomic'
            self.wait = '--wait'
        else:
            self.atomic = ''
            self.wait = ''
        if 'timeout' in config:
            self.timeout = '--timeout {}'.format(config['timeout'])
        else:
            self.timeout = ''
        self.tpl = tpl
        self.tmp = tmp
        self.values = 'v.yaml'

        v = '{} version --short'.format(config['command'])
        v = subprocess.run(
            shlex.split(v),
            check=True, text=True, capture_output=True
        )
        self.version = v.stdout.strip()
        v = '{} diff version'.format(config['command'])
        v = subprocess.run(
            shlex.split(v),
            check=True, text=True, capture_output=True
        )
        self.version_diff = v.stdout.strip()

    def _render_v(self, env, values):
        with open(values, 'r') as f:
            v = self.tpl.render(f.read(), env)
        with open('{}/{}'.format(self.tmp, self.values), 'w') as f:
            f.write(v)

    def _run(self, env, args, hide_stdout=False):
        c = self.tpl.render(self.cmd, env).strip()
        c = '{} {}'.format(c, args)
        if hide_stdout:
            p = subprocess.Popen(
                shlex.split(c),
                stdout=subprocess.DEVNULL,
                stderr=subprocess.PIPE,
                encoding='utf-8', errors='ignore',
                bufsize=0
            )
            prev_ln = ''
            nl = True
            for ln in iter(p.stderr.readline, ''):
                ln = ln.rstrip()
                if '[debug]' in ln:
                    ln = ln.split(': ', 1)[1]
                    ln_short = ln.split('[debug]')[1]
                else:
                    ln_short = ln
                if ln_short == prev_ln:
                    print('.', end='', flush=True)
                    nl = False
                else:
                    if not nl:
                        print()
                        nl = True
                    print(ln)
                prev_ln = ln_short
            p.stderr.close()
            p.wait()
            if p.returncode != 0:
                raise subprocess.SubprocessError(
                    "Command '{}' terminated with return code {}"
                    .format(c, p.returncode)
                )
        else:
            subprocess.run(shlex.split(c), check=True, text=True)

    def template(self, release, env, values):
        self._render_v(env, values)
        self._run(
            env,
            'template -f {}/{} {} .'
            .format(self.tmp, self.values, release)
        )

    def debug(self, release, env, values):
        self._render_v(env, values)
        self._run(
            env,
            'upgrade -i --debug --dry-run -f {}/{} {} .'
            .format(self.tmp, self.values, release)
        )

    def diff(self, release, env, values):
        self._render_v(env, values)
        self._run(
            env,
            'diff upgrade -f {}/{} {} .'
            .format(self.tmp, self.values, release)
        )

    def deploy(self, release, env, values):
        self._render_v(env, values)
        self._run(
            env,
            'upgrade --debug -i -f {}/{} {} {} {} .'
            .format(self.tmp,
                    self.values,
                    self.atomic,
                    self.timeout,
                    release),
            True
        )

    def status(self, release, env):
        self._run(
            env,
            'list -f "^{}$"'.format(release)
        )

    def delete(self, release, env):
        self._run(
            env,
            'uninstall --debug {} {} {}'
            .format(self.wait, self.timeout, release),
            True
        )



import jinja2

version = 'jinja2 {}'.format(jinja2.__version__)
regex = None


def render(tpl, env):
    # rise an error if tpl contains variables which are not in env
    j2env = jinja2.Environment(undefined=jinja2.StrictUndefined)

    return j2env.from_string(tpl, env).render()



#!/usr/bin/env python3

import yaml
from threading import Thread
from pprint import pformat
import os
import itertools


def flatten_matrix(matrix):
    flat_matrix = []
    item_list = []
    for var_name, var_value_list in matrix.items():
        if type(var_value_list) is not list:
            var_value_list = [var_value_list]
        item = []
        for var_value in var_value_list:
            item.append([var_name, var_value])
        item_list.append(item)
    for item in list(itertools.product(*item_list)):
        flat_item = {}
        for (var_name, var_value) in item:
            flat_item[var_name] = var_value
        flat_matrix.append(flat_item)
    return flat_matrix


# Count the variables in a release
def count_vars(data, tpl):
    c = 0
    for var_name, var_value in data.items():
        if var_name != '__count__':
            if tpl.var_re:
                c = c + len(tpl.var_re.findall(var_value))
            else:
                for k in data.keys():
                    c = c + var_value.count(k)
    if c == 0 and not tpl.var_re:
        for var_name, var_value in data.items():
            if var_name != '__count__':
                try:
                    tpl.render(var_value, {})
                except Exception:
                    c = 999
                    break
    return c


def solve_vars(in_data, out_data, tpl):
    in_data['__count__'] = count_vars(in_data, tpl)
    while in_data['__count__'] > 0:
        tmp_data = {'__count__': 0}
        c = in_data['__count__']
        del(in_data['__count__'])
        err = None
        for var_name, var_value in in_data.items():
            try:
                tmp_data[var_name] = tpl.render(var_value, in_data).strip()
            except Exception as e:
                err = e
                tmp_data[var_name] = var_value
        tmp_data['__count__'] = count_vars(tmp_data, tpl)
        if c == tmp_data['__count__']:
            out_data['e']['last'] = in_data
            out_data['e']['err'] = err
            out_data['e']['count'] = c
            return
        in_data = tmp_data
    out_data['d'] = dict(in_data)


def get_releases(tpl, tmp, config):
    cache = '{}/releases.{}.yaml'.format(tmp, config['.hash'])
    if os.path.isfile(cache):
        with open(cache, 'r') as f:
            out = yaml.safe_load(f.read())
    else:
        # Flatten releases
        out = []
        for release_name, data in config['releases'].items():
            release_template = {'__name__': release_name}
            if 'variables' in config:
                for var_name, var_value in config['variables'].items():
                    release_template[var_name] = var_value
            if 'variables' in data:
                for var_name, var_value in data['variables'].items():
                    release_template[var_name] = var_value
            if 'matrix' in data:
                for matrix in data['matrix']:
                    flat_matrix = flatten_matrix(matrix)
                    for release_matrix in flat_matrix:
                        out.append({**release_template, **release_matrix})
            else:
                out.append(release_template)

        # Remove empty vars
        for release in out:
            to_del = []
            for var_name, var_value in release.items():
                if not var_value:
                    to_del.append(var_name)
            for var_name in to_del:
                del(release[var_name])

        # Solve vars, each release in a thread
        _releases = []
        threads = []
        for release in out:
            x = {
                't': None,    # thread
                'r': {        # solved vars release
                    'd': {},  # data
                    'e': {}   # err
                }
            }
            t = Thread(target=solve_vars,
                       args=(release, x['r'], tpl),
                       daemon=True)
            x['t'] = t
            threads.append(x)
            t.start()
        for x in threads:
            x['t'].join()
            if len(x['r']['e']) == 0:
                _releases.append(x['r']['d'])
            else:
                msg = ''
                msg = msg + 'Loop or template error\n'
                msg = msg + 'Latest data:\n'
                msg = msg + pformat(x['r']['e']['last']) + '\n'
                msg = msg + 'Error: {}\n'.format(x['r']['e']['err'])
                raise RuntimeError(msg)
        out = _releases

        # Cleanup variables staring with _
        # Check for duplicate names
        _releases = {}
        for release in out:
            n = release['__name__']
            if n not in _releases:
                to_del = []
                for k in release:
                    if k[0] == '_':
                        to_del.append(k)
                for k in to_del:
                    del(release[k])
                _releases[n] = release
            else:
                raise RuntimeError(
                    'Duplicate release name: {}\n{}\n{}'.format(
                        n, _releases[n], release)
                )
        out = _releases

        with open(cache, 'w') as f:
            f.write(yaml.safe_dump(out))

    return out



import subprocess
import shlex
import os
import sys
import re
import importlib
import requests


class Template():

    def __init__(self, config, tmp):
        self.tmp = tmp
        self.cmd = None
        self.module = None
        if 'module' in config:
            module = config['module']
            if not module.endswith('.py'):
                self.module = importlib.import_module(
                    'timona.template.module_{}'.format(module)
                )
            else:
                sys.path = [tmp] + sys.path
                if module.startswith('http://') or \
                        module.startswith('https://'):
                    r = requests.get(module, timeout=(0.5, 5))
                    r.raise_for_status()
                    m = r.text
                else:
                    with open(module) as f:
                        m = f.read()
                with open('{}/module_template.py'.format(tmp), 'w') as f:
                    f.write(m)
                self.module = importlib.import_module('module_template')
            v = 'module: {}'.format(self.module.version)
            var_re = self.module.regex
        elif 'command' in config:
            self.cmd = config['command']['render']
            p = subprocess.run(
                shlex.split(config['command']['version']),
                capture_output=True, check=True, text=True
            )
            v = p.stdout.strip()
            v = 'command: {}'.format(v)
            self.stderr = config['command'].get('stderr')
            var_re = config['command'].get('regex')
        self.version = v
        self.var_re = var_re
        if self.var_re:
            self.var_re = re.compile(self.var_re)

    def render(self, tpl, env):
        if self.module:
            return self.module.render(tpl, {**env, **os.environ})
        elif self.cmd:
            try:
                p = subprocess.run(
                    shlex.split(self.cmd),
                    input=tpl, env={**env, **os.environ},
                    capture_output=True, check=True, text=True,
                )
            except subprocess.CalledProcessError as e:
                raise subprocess.SubprocessError(e.stderr) from e
            if p.stderr.strip():
                if self.stderr == 'print':
                    print(p.stderr, file=sys.stderr)
                elif self.stderr == 'error':
                    raise subprocess.SubprocessError(
                        'Render error:\n{}'.format(p.stderr)
                    )
            return p.stdout
